/*
 * sbx81cfc960.c
 *
 * Copyright 2012 Allied Telesis Labs
 *
 * Based on p2041rdb.c
 *
 * Copyright 2011 Freescale Semiconductor, Inc.
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#include <common.h>
#include <malloc.h>
#include <command.h>
#include <netdev.h>
#include <linux/compiler.h>
#include <asm/mmu.h>
#include <asm/processor.h>
#include <asm/cache.h>
#include <asm/immap_85xx.h>
#include <asm/fsl_law.h>
#include <asm/fsl_serdes.h>
#include <asm/fsl_portals.h>
#include <asm/fsl_liodn.h>
#include <fm_eth.h>
#include <idprom.h>
#include <flash.h>
#include <fdt.h>
#include <libfdt.h>
#ifdef CONFIG_SPIFLASH
#include <spi_flash.h>
#endif
#include "../SBx8100/mpi.h"
#ifdef CONFIG_USB_STORAGE
#include <usb.h>
#endif
#include <98dx160.h>
#include "../SBx8100/dx160.h"

extern void pci_of_setup(void *blob, bd_t *bd);

DECLARE_GLOBAL_DATA_PTR;

int checkboard(void)
{
	struct cpu_type *cpu = gd->arch.cpu;
	ccsr_gur_t *gur = (void *)CONFIG_SYS_MPC85xx_GUTS_ADDR;
	unsigned int i;

	printf("Board: SBx81CFC960 (%s), ", cpu->name);

#ifdef CONFIG_PHYS_64BIT
	puts("36-bit Addressing\n");
#endif

	/*
	 * Display the RCW, so that no one gets confused as to what RCW
	 * we're actually using for this boot.
	 */
	puts("Reset Configuration Word (RCW):");
	for (i = 0; i < ARRAY_SIZE(gur->rcwsr); i++) {
		u32 rcw = in_be32(&gur->rcwsr[i]);

		if ((i % 4) == 0)
			printf("\n       %08x:", i * 4);
		printf(" %08x", rcw);
	}
	puts("\n");

	return 0;
}

int board_early_init_f(void)
{
	ccsr_gur_t *gur = (void *)(CONFIG_SYS_MPC85xx_GUTS_ADDR);

	/* board only uses the DDR_MCK0/1, so disable the DDR_MCK2/3 */
	setbits_be32(&gur->ddrclkdr, 0x000f000f);

	return 0;
}

int board_early_init_r(void)
{
	const unsigned int flashbase = CONFIG_SYS_FLASH_BASE;
	const u8 flash_esel = find_tlb_idx((void *)flashbase, 1);

	/*
	 * Remap Boot flash + PROMJET region to caching-inhibited
	 * so that flash can be erased properly.
	 */

	/* Flush d-cache and invalidate i-cache of any FLASH data */
	flush_dcache();
	invalidate_icache();

	/* invalidate existing TLB entry for flash */
	disable_tlb(flash_esel);

	set_tlb(1, flashbase, CONFIG_SYS_FLASH_BASE_PHYS,
			MAS3_SX|MAS3_SW|MAS3_SR, MAS2_I|MAS2_G,
			0, flash_esel, BOOKE_PAGESZ_256M, 1);

	set_liodns();
	setup_portals();

	return 0;
}

unsigned long get_board_sys_clk(unsigned long dummy)
{
	return 100000000;
}

static int get_slot_id(void)
{
	int slot = in8(FPGA_BASE + 0x04);

	/* b4:0 is slotid - with b4 inverted */
	slot &= 0x1f;
	slot ^= 0x10;

	return slot;
}

static void setenv_slot_id(void)
{
	int slot;
	char buf[16];

	if ((slot = get_slot_id()) < 0) {
		fprintf(stderr, "Error: could not determine slot number\n");
		slot = 0;
	}

	sprintf(buf, "%d", slot);
	setenv("slot_id", buf);
}

static void setenv_eth_addrs(void)
{
	uchar slot_id = simple_strtoul(getenv("slot_id"), NULL, 10);
	uchar unused_eth_mac[6] = {0x00, 0x00, 0x00, 0x00, 0x00, 0x00};
	uchar bp_eth_mac[6] = {0x02, 0x00, 0xcd, 0x00, 0x00, slot_id};

#ifdef CONFIG_DIAG_IDPROM
	idprom_setmacaddr(0);
#endif
	eth_setenv_enetaddr("eth1addr", unused_eth_mac);
	eth_setenv_enetaddr("eth2addr", unused_eth_mac);
#ifdef CONFIG_DIAG_IDPROM
	idprom_setmacaddr(3);
#endif
	eth_setenv_enetaddr("eth4addr", bp_eth_mac);
}

int misc_init_r(void)
{
	mpi_board_init();

	setenv_slot_id();
	setenv_eth_addrs();

	/* Enable loading of the FMAN ucode from SPI */
	if(mpi_board_attached())
		mpi_spi_flash_select(1,0);

#ifdef MTDPARTS_MTDOOPS
	setenv("mtdoops", MTDPARTS_MTDOOPS);
#endif

	return 0;
}

/* This is run after timer_interrupt is enabled */
int last_stage_init(void)
{
#ifdef CONFIG_USB_STORAGE
	usb_init();
#endif
	return 0;
}

static int fdt_fixup_flash_partitions(void *blob, int parent)
{
	int err, node, i;
	u32 tmp[2];
	struct {
		char *name;
		u32 addr;
		u32 size;
	} paritions[] = {
		{"errlog",     0xff00000,         0x100000},
		{"user", CONFIG_UBOOT_SIZE+CONFIG_ENV_SECT_SIZE, 0xfde0000},
		{"u-boot-env", CONFIG_UBOOT_SIZE, CONFIG_ENV_SECT_SIZE},
		{"u-boot",     0,                 CONFIG_UBOOT_SIZE},
	};

	for (i = 0; i < 4; i++) {
		char part_name[32];
		snprintf(part_name, sizeof(part_name), "partition@%s",
			paritions[i].name);

		node = fdt_add_subnode(blob, parent, part_name);
		if (node < 0) {
			printf("%s: Failed to create FDT node %s %s\n",
				__func__, paritions[i].name,
				fdt_strerror(node));
			return -1;
		}

		err = fdt_setprop(blob, node, "label", paritions[i].name,
				strlen(paritions[i].name)+1);
		if (err < 0) {
			printf("%s: Failed to set \"label\" property %s %s\n",
				__func__, paritions[i].name,
				fdt_strerror(err));
			return -1;
		}

		if (paritions[i].addr == 0) {
			err = fdt_setprop(blob, node, "read-only", NULL, 0);
			if (err < 0) {
				printf("%s: Failed to set \"read-only\" "
					"property %s %s\n",
					__func__, paritions[i].name,
					fdt_strerror(err));
				return -1;
			}
		}

		tmp[0] = cpu_to_fdt32(paritions[i].addr);
		tmp[1] = cpu_to_fdt32(paritions[i].size);
		err = fdt_setprop(blob, node, "reg", tmp, sizeof(tmp));
		if (err < 0) {
			printf("%s: Failed to set \"reg\" property %s %s\n",
				__func__, paritions[i].name,
				fdt_strerror(err));
			return -1;
		}
	}

	return 0;
}

/**
 * Determines whether this fixup is required. If /flash@1
 * is present then the workaround is needed.
 */
static int fdt_fixup_flash_required(void *blob)
{
	int node;

	node = fdt_path_offset(blob, "/localbus/flash@1,0");
	if (node < 0) {
		return 0;
	}

	return 1;
}

static int fdt_fixup_flash(void *blob)
{
	int err, node;
	u32 tmp[CONFIG_SYS_MAX_FLASH_BANKS*3];
	int i;
	int bank_width = cpu_to_fdt32(2);
	int device_width = cpu_to_fdt32(2);

	if (!fdt_fixup_flash_required(blob)) {
		printf("%s: Fixup not required\n", __func__);
		return 0;
	}

	// Delete the old flash FDT entries
	node = fdt_path_offset(blob, "/localbus/flash@0,0");
	if (node < 0) {
		printf("%s: flash@0,0 : %s\n", __func__, fdt_strerror(node));
		return -1;
	}

	err =  fdt_del_node(blob, node);
	if (err < 0) {
		printf("%s: flash@0,0 %s\n", __func__, fdt_strerror(err));
		return -1;
	}

	node = fdt_path_offset(blob, "/localbus/flash@1,0");
	if (node < 0) {
		printf("%s: flash@1,0 : %s\n", __func__, fdt_strerror(node));
		return -1;
	}

	err =  fdt_del_node(blob, node);
	if (err < 0) {
		printf("%s: flash@1,0 %s\n", __func__, fdt_strerror(err));
		return -1;
	}

	// Re-create flash@0,0
	node = fdt_path_offset(blob, "/localbus");
	if (node < 0) {
		printf("%s: Could not find \"localbus\" %s",
			__func__, fdt_strerror(node));
		return -1;
	}

	node = fdt_add_subnode(blob, node, "flash@0,0");
	if (node < 0) {
		printf("%s: Could not create \"flash@0,0\" %s\n",
			__func__, fdt_strerror(node));
		return -1;
	}

	err = fdt_setprop(blob, node, "compatible", "cfi-flash",
			sizeof("cfi-flash"));
	if (err < 0) {
		printf("%s: Could not set property \"compatible\" %s",
			__func__, fdt_strerror(err));
		return -1;
	}

	for (i = 0; i < CONFIG_SYS_MAX_FLASH_BANKS; i++) {
		if (flash_info[i].flash_id == FLASH_UNKNOWN)
			continue;

		tmp[i*3] =  cpu_to_fdt32(i);
		tmp[(i*3)+1] = 0;
		tmp[(i*3)+2] = cpu_to_fdt32(flash_info[i].size);
	}

	err = fdt_setprop(blob, node, "reg", tmp, sizeof(tmp));
	if (err < 0) {
		printf("%s: Could not set property \"reg\" %s",
			__func__, fdt_strerror(err));
		return -1;
	}

	err = fdt_setprop(blob, node, "bank-width", &bank_width,
		sizeof(bank_width));
	if (err < 0) {
		printf("%s: Could not set property \"bank-width\" %s",
			__func__, fdt_strerror(err));
		return -1;
	}

	err = fdt_setprop(blob, node, "device-width", &device_width,
		sizeof(device_width));
	if (err < 0) {
		printf("%s: Could not set property \"device-width\" %s",
			__func__, fdt_strerror(err));
		return -1;
	}

	// Create partition subnodes
	err = fdt_fixup_flash_partitions(blob, node);
	if (err < 0) {
		printf("%s: Could not setup partitions",
			__func__);
		return -1;
	}

	return 0;
}

void ft_board_setup(void *blob, bd_t *bd)
{
	phys_addr_t base;
	phys_size_t size;

	ft_cpu_setup(blob, bd);

	base = getenv_bootm_low();
	size = getenv_bootm_size();

	fdt_fixup_memory(blob, (u64)base, (u64)size);

#if defined(CONFIG_HAS_FSL_DR_USB) || defined(CONFIG_HAS_FSL_MPH_USB)
	fdt_fixup_dr_usb(blob, bd);
#endif

#ifdef CONFIG_PCI
	pci_of_setup(blob, bd);
#endif

	fdt_fixup_liodn(blob);
#ifdef CONFIG_SYS_DPAA_FMAN
	fdt_fixup_fman_ethernet(blob);
#endif
	fdt_fixup_flash(blob);
}

void board_reset(void)
{
	out8(FPGA_BASE + 0x24, 0xa5);
}

#ifdef CONFIG_SPIFLASH
void *spi_firmware_get(void)
{
	int ret;
	void *blob;
	struct spi_flash *flash;

	flash = spi_flash_probe(CONFIG_ENV_SPI_BUS, CONFIG_ENV_SPI_CS,
				CONFIG_ENV_SPI_MAX_HZ, CONFIG_ENV_SPI_MODE);
	if (!flash) {
		printf("SF: probe failed\n");
		return NULL;
	} else {
		blob = calloc(1, CONFIG_FIRMWARE_BLOB_SIZE);
		ret = spi_flash_read(flash, CONFIG_FIRMWARE_BLOB_ADDR,
				CONFIG_FIRMWARE_BLOB_SIZE, blob);
		if (ret)
			printf("SF: read failed\n");
		spi_flash_free(flash);
	}

	return blob;
}
#endif

/* atl_get_env_addr()
 *
 * Get location of U-Boot environement. ENV code was hacked a little to make
 * this work. Doesn't do anything special for the CFC960.
 */
u32 atl_get_env_addr(void)
{
	return CONFIG_ENV_ADDR;
}

/* atl_get_offset_from_base()
 *
 * Not used by this product
 */
u32 atl_get_offset_from_base(void)
{
	return 0x0;
}

int cfc960_dx160_setup(void)
{
	int i;

	/* Ensure DX160 is in reset */
	out8(FPGA_BASE + 0x20, in8(FPGA_BASE + 0x20) | (1 << 2));
	udelay(100);

	/* De-assert SKIP_INITn (ie de-assert SKIP_INIT) */
	out8(FPGA_BASE + 0x11, in8(FPGA_BASE + 0x11) & ~(1 << 5));
	udelay(100);

	/* Bring DX160 out of reset */
	out8(FPGA_BASE + 0x20, in8(FPGA_BASE + 0x20) & ~(1 << 2));

	/* Wait for DEV_INIT_DONE */
	for (i = 0; i < 10; i++) {
		if ((in8 (FPGA_BASE + 0x10) & (1 << 4)) != 0);
			break;
		udelay(100);
	}

	/* Configure DX160 for SBx81CFC operation */
	dx160_config("FSL_MDIO0", CONFIG_98DX160_ADDR, cfc_dx160_config);

	/* Assert EN_DEV to enable operation */
	out8(FPGA_BASE + 0x11, in8(FPGA_BASE + 0x11) | (1 << 4));

	/* Dumping registers brings link up. Not sure why. */
	dx160_read_all();

	return 0;
}

void try_bootp_boot(void)
{
	char buf[32];

	/* Get the address we want to use */
	uchar eth_mac[6];
	eth_getenv_enetaddr("eth4addr", eth_mac);
	eth_setenv_enetaddr("eth1addr", eth_mac);

	cfc960_dx160_setup();
	//start with eth1
	setenv("ethact", "FM1@DTSEC5");
	//only try for one retry cycle.
	setenv("netretry", "no");

	if (run_command("bootp", 0) >= 0) {
		sprintf(buf, "awbootm %s", getenv("fileaddr"));
		run_command(buf, 0);
	}
}
