#include <common.h>
#include <sha256.h>
#include <password.h>
#include <api_public.h>
#include <bootmenu.h>

#ifdef CONFIG_USE_PASSWORD

DECLARE_GLOBAL_DATA_PTR;
/*
 *	sha256()
 *
 *	Apply SHA256 to input
 */
static void
sha256(uint8_t * input, uint32_t length, uint8_t digest[32])
{
	sha256_context ctx;
	sha256_starts(&ctx);
	sha256_update(&ctx, input, length);
	sha256_finish(&ctx, digest);
}

/*
 *	charify_digest()
 *
 *	Represent raw hex as ASCII
 */
static void
charify_digest(uint8_t digest[32], char *charified)
{

	int i;
	char *c = charified;

	for (i = 0; i < 32; i++) {
		sprintf(c, "%02x", digest[i]);
		c += 2;
	}
}
/*
 *	process_password()
 *
 *	Receive a password as input
 *	Hash it
 *	Represent the 32 bytes digest using a 256 bytes ASCII string
 */
static void
process_password(char *password, size_t password_length, char *hash)
{
	uint8_t digest[32];
	char charified[HASH_SIZE];
	uint8_t *input = (uint8_t *) password;
	uint32_t input_length = (uint32_t) password_length;

	sha256(input, input_length, digest);

	charify_digest(digest, charified);

	memcpy(hash, charified, HASH_SIZE);
	hash[HASH_SIZE] = '\0';
}
/*
 *	security_level_get()
 *
 *	Return the current security level as a string
 *	SEC_LEVELx_DESCRIPTION
 */
char *
security_level_get()
{
	char *ch = NULL;

	ch = getenv(SECURITY_LEVEL_ENV_VAR);
	if(!ch)
	{
		return "";
	}

	if(!(strcmp(ch, SEC_LEVEL1)))
		return SEC_LEVEL1_DESCRIPTION;
	else if(!(strcmp(ch, SEC_LEVEL2)))
		return SEC_LEVEL2_DESCRIPTION;
	else if(!(strcmp(ch, SEC_LEVEL3)))
		return SEC_LEVEL3_DESCRIPTION;

	return ""; /* not supported */
}
/*
 *	security_level_check()
 *
 *	Check for a match in security level
 *	Return 1 if it matches, 0 otherwise
 */
int
security_level_check(char *sec_level)
{
	int result = 0;
	char *ch = NULL;

	ch = getenv(SECURITY_LEVEL_ENV_VAR);
	if(!ch)
	{
		return 1;
	}
	if(!(strcmp(ch, sec_level)))
		result = 1;

	return result;
}

/*
 *	security_level_set()
 *
 *	Set the current security level
 *	Return 0 if successful , 1 otherwise
 */
int
security_level_set(char *new_security_level)
{
	setenv(SECURITY_LEVEL_ENV_VAR, new_security_level);
	if (saveenv()? 1 : 0)
		return 1;
	return 0;
}

/* protection_set()
 *
 *	Set / change password protection
 *	Return 0 if successful , 1 otherwise
 */
int
protection_set(char *password, size_t password_length)
{
	char hash[HASH_STR_SIZE];
	process_password(password, password_length, hash);
	setenv(PASSWORD_ENV_VAR, hash);
	return (saveenv()? 1 : 0);
}

/*
 *	lock_status_init()
 *
 *	Init lock status based on sec level
 *	Must be called at start up
 *
 */
int
lock_status_init()
{
	char *l = NULL;
	l = getenv(SECURITY_LEVEL_ENV_VAR);
	if(l == NULL) /* this was never set so init all */
	{
		SET_STATE_UNLOCKED;
		/* if relocated to RAM */
		if (gd->flags & GD_FLG_RELOC) /* setting env before relocation is not safe and somewhat useless */
		{
			setenv(SECURITY_LEVEL_ENV_VAR, SEC_LEVEL1);
			saveenv();
		}
	}
	else if (!(strcmp(l, SEC_LEVEL1)))
	{
		SET_STATE_UNLOCKED;//lock_state = UNLOCKED;  /* level 1 does not required any password  */
	}
	else
	{
		SET_STATE_LOCKED;  /* all other cases, we need the password */
	}
	return 0;
}

/*
 *	password_verify()
 *
 *	Receive a password as input
 *	Process it
 *	Determine if the input and the stored password match
 */
static int
password_verify(char *password, size_t length)
{

	char *original = NULL;
	char orig_f[HASH_STR_SIZE];
	char hash[HASH_STR_SIZE];

	if (gd->flags & GD_FLG_ENV_READY) {
		original = getenv(PASSWORD_ENV_VAR);
	} else {
		getenv_f(PASSWORD_ENV_VAR, orig_f, sizeof(orig_f));
		original = orig_f;
	}
	process_password(password, length, hash);
	return strncmp(hash, original, HASH_SIZE);
}

/*
 *	unlock_secure()
 *
 *	If the password supplied match, unlock the lock
 */
password_state_t
unlock_secure(char *password, size_t length)
{
	if(!password_verify(password, length))
	{
		SET_STATE_UNLOCKED;
		return UNLOCKED;
	}
	else
	{
		SET_STATE_LOCKED;
		return LOCKED;
	}
}

/* prompt for password if needed
 * return LOCKED/UNLOCKED status*/
int password_protected_section(int *quit)
{
	int result = trap_loop();
	if (result == PROMPT_TIMEOUT)
	{
		if(quit != NULL)
			*quit = 1;
	}
	return result;
}

/*
 *	trap_loop ()
 *
 *	Lock console execution
 */
password_state_t
trap_loop()
{
	char buffer[BUFFER_MAX_LEN] = "\0";
	int i;

	if (PASSWORD_STATE_LOCKED != UNLOCKED)
	{
		memset(buffer, 0, BUFFER_MAX_LEN);
		bootmenu_printf( "\r\n");
		bootmenu_printf("  Authentication required \r\n");

		bootmenu_printf("  Please enter password:");

		i = bootmenu_get_string(buffer, BUFFER_MAX_LEN, ECHO_PASSWORD, TIMEOUT_PASSWORD_PROMPT);

		if (i == PROMPT_TIMEOUT)
		{
			bootmenu_printf("\r\n\r\n");
			return PROMPT_TIMEOUT;
		}

		if ((i >= PASSWORD_MIN_LEN)  &&
			(i <= PASSWORD_MAX_LEN)  &&
			unlock_secure(buffer, (size_t)i) == UNLOCKED)
		{
			bootmenu_printf("\r\n\r\n");
			return UNLOCKED;
		}

		udelay (3000000); // 3s /* delay to discourage Brute Force Attack */

		bootmenu_printf( "\r\n");
		bootmenu_printf("  Incorrect password \r\n");
		bootmenu_printf("\n\n");

		udelay (3000000); // 3s

		return LOCKED;
	}
	return UNLOCKED;
}
#endif
