/*
 * Copyright 2011 Allied Telesis Labs
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#include <common.h>
#include <command.h>
#include <asm/addrspace.h>
#include <epi3.h>
#include <bootmenu.h>
#include <i2c.h>
#include <asm/io.h>
#ifdef CONFIG_USB_STORAGE
#include <usb.h>
#endif


#define ALP_CLOCK				25000000
#define CC_CORECTL_UART_CLKOV_MASK    0x1
#define CC_CAP_UINTCLK			0x00000008  /* UARTs are driven by internal divided clock */
#define CORE_CONTROL			0x18000008
#define CHIP_STATUS				0x1800002c
#define CHIPSTAT_PO_MASK		0x1 /* Package Option */
#define CLKD_UART				0x000000ff
#define FLASHSTR_CONFIG			0x1800018c
#define GPIOOUT_REG				0x18000064
#define GPIOOUTEN_REG			0x18000068
#define GPIOCONTROL_REG			0x1800006c
#define GPIOINTMASK_REG			0x18000074
#define GPIOINTPOLARITY_REG		0x18000070
#define PFLASH_WAITCOUNT		0x1800012c
#define PFLASH_WAITCOUNT_PF1	0x18000130
#define SERIAL_IO_SEL_REG		0x180002f4
#define PLL_CONTROLLER_ADDR		0x18000660
#define PLL_CONTROLLER_DATA		0x18000664
#define WDOG_COUNTER			0x18000080
#define X310_CORE_CONTROL_REG	0x18000008
#define X310_CAP_REG			0x18000004
#define X310_CLOCK_DIV_REG		0x180000A4

/* GPIO Bits */
#define RESET_ALL_PHY_N		(1<<0)
#define RESET_MAC_N			(1<<1)
#define GPIO_MANUAL_RST		(1<<2)
#define OCTAL_PHY_INT_N		(1<<3)
#define GPIO4_I2C_SCL		(1<<4)
#define RESET_POE_MEZZ_N	(1<<8)
#define RESET_I2C_GPIO_N	(1<<9)
#define RESET_I2C_MUX_N		(1<<10)
#define GPIO_I2C_INT_N		(1<<11)
#define FAN_INT_N			(1<<12)
#define USB_PWR_EN			(1<<13)
#define GREEN_BUTTON		(1<<14)
#define POE_PORT_DIS_N		(1<<15)

extern uint x310_get_cpu_cap(void);
extern uint x310_get_chipid(void);
int cfi_flash_num_flash_banks;

ulong flash_get_size (phys_addr_t base, int banknum);
extern flash_info_t  flash_info[];


int bootflash_attached = 0;

/* Write to a CPU register. */
inline static void reg_write(u32 addr, u32 data)
{
    writel(data, KSEG1ADDR(addr));
}

/* Read a CPU register */
inline static u32 reg_read(u32 addr)
{
    return readl(KSEG1ADDR(addr));
}

/* atl_bootrom_board_attached()
 *
 * used by bootmenu to determine whether or not to print
 * option '5' (EPROM) in upgrade bootloader menu
 */

int atl_bootrom_board_attached(void)
{
	return bootflash_attached;
}

/* atl_get_offset_from_base()
 *
 * not needed on this product
 */

u32 atl_get_offset_from_base(void)
{
	return 0;
}

/* atl_get_env_addr()
 *
 * return address of env for places where a macro can't be used.
 */

u32 atl_get_env_addr(void)
{
	return CONFIG_ENV_ADDR;
}

phys_size_t initdram(int board_type)
{
	/* Sdram is setup by assembler code */
	/* If memory could be changed, we should return the true value here */
	return 512 * 1024 * 1024; /* should be 512M must fix */
}

u32 x310_serial_base_clock(void)
{

	uint32_t  w, clock = ALP_CLOCK;

	w = reg_read(X310_CORE_CONTROL_REG);
	if ((w & CC_CORECTL_UART_CLKOV_MASK)) {
		/* internal ALP clock */
		clock = ALP_CLOCK;
	} else {
		w = reg_read(X310_CAP_REG);
		/* Check the clock is internal or external */
		if ((w & CC_CAP_UINTCLK)) {
			/* Internal ALP clock */
			w = reg_read(X310_CLOCK_DIV_REG);
			w = w & CLKD_UART;
			clock = (clock / w);
		}
	}

	return clock;
}


u32 x310_system_counter_freq(void)
{
	u32  w;
	u32 cpu_freq;

	w = reg_read(CHIP_STATUS);
	if (w & CHIPSTAT_PO_MASK) {
		cpu_freq = 400000000;
	} else {
		// return (60000000); /* shorten the delay */
		/* Get N divider to determine CPU clock */
		reg_write(PLL_CONTROLLER_ADDR, 4);
		w = reg_read(PLL_CONTROLLER_DATA);
		w = ( w & 0x00000fff) >> 3;

		/* Fixed reference clock 25MHz and m = 2 */
		cpu_freq = w * 25000000 / 2;
	}
	return cpu_freq;
}

/* It is possible for a device to be stuck in the middle of a transfer when the
 * unit is rebooted. In this scenario we have observed the i2c to lock up.
 * The fix is to simulate more than a byte's worth of clock signals to trigger
 * the hung up device to finish its transfer and free up the data line.
 */
static int i2c_lockup_clear(void)
{
	uint io_sel;
	uint gpio_out;
	uint gpio_out_en;
	int i;

	/* Turn the pins used for i2c back into GPIOs */
	io_sel = reg_read(SERIAL_IO_SEL_REG);
	reg_write(SERIAL_IO_SEL_REG, (io_sel & ~1));

	/* Enable GPIO4 so we can toggle it as the clock line */
	gpio_out_en = reg_read(GPIOOUTEN_REG);
	reg_write(GPIOOUTEN_REG, (gpio_out_en | GPIO4_I2C_SCL));

	/* Send a bunch of simulated clock signals */
	gpio_out = reg_read(GPIOOUT_REG);
	for (i = 0; i < 10; i++) {
		reg_write(GPIOOUT_REG, (gpio_out & ~GPIO4_I2C_SCL));
		udelay(10);
		reg_write(GPIOOUT_REG, (gpio_out | GPIO4_I2C_SCL));
		udelay(10);
	}

	/* Disable GPIO4 as a GPIO*/
	reg_write(GPIOOUTEN_REG, gpio_out_en);

	/* Set the i2c pins back to i2c mode */
	reg_write(SERIAL_IO_SEL_REG, (io_sel | 1));

	return 0;
}

int checkboard (void)
{
	uint chipid;

	printf("Board: x310\n");

	chipid = x310_get_chipid();
	printf ("Chip revision %04x device %04x, ", (chipid >> 16), (chipid & 0xffff));
	printf("CPU Speed %d MHz\n", (int)(x310_system_counter_freq() /1000/1000));

	/* Clear any i2c device holding the data line due to reboot during a data
	 * transfer
	 */
	i2c_lockup_clear();

	return 0;
}

int board_eth_init(bd_t *bis)
{
#ifdef CONFIG_USB_HOST_ETHER
	usb_host_eth_scan(1);
#endif

	return 0;
}

int board_early_init_f(void)
{
	/* Set up the GPIOs */
	/* Clear all bits as not using IO pins in special cores */
	reg_write(GPIOCONTROL_REG, 0x0000);

	/* Set GPIO pins. Take switch chip out of reset. */
	reg_write(GPIOOUT_REG, RESET_POE_MEZZ_N | RESET_I2C_GPIO_N |
	                       RESET_I2C_MUX_N | USB_PWR_EN |
	                       POE_PORT_DIS_N | RESET_MAC_N);

	/* Enable GPIO pins as outputs = 0xA703 */
	reg_write(GPIOOUTEN_REG, RESET_ALL_PHY_N | RESET_MAC_N |
	                         RESET_POE_MEZZ_N | RESET_I2C_GPIO_N |
	                         RESET_I2C_MUX_N | USB_PWR_EN |
	                         POE_PORT_DIS_N);

	/* setup the interrupt inputs, trigger on a high to low signal change */
	reg_write(GPIOINTMASK_REG, 0x00000000);

	/* trigger on a high to low signal change =  0x5808 */
	reg_write(GPIOINTPOLARITY_REG, OCTAL_PHY_INT_N | FAN_INT_N |
	                               GPIO_I2C_INT_N | GREEN_BUTTON);

	return 0;
}

int misc_init_r (void)
{
	setenv("ethact", "switch");
#ifdef CONFIG_DIAG_EPI3
	epi3_setmacaddr();
#endif

#if defined(CONFIG_USB_STORAGE) || defined(CONFIG_USB_HOST_ETHER)
	/* wait for the USB power lane to be stable */
	udelay(100*1000);
	usb_init();
#endif

	/* turn off watchdog timer on RTC device */
	i2c_reg_write(0x68, 0xc, 0x0);

	/* light up 7-segment LED */
	i2c_reg_write(0x20, 0x1, 0x0);  /* output value */
	i2c_reg_write(0x20, 0x3, 0x0);  /* control */

#ifdef MTDPARTS_MTDOOPS
	setenv("mtdoops", MTDPARTS_MTDOOPS);
#endif
	return 0;
}

u32 atl_get_jffs2_part_size(void)
{
	u32 partition_size = 0;
	int i;

	for (i = 0; i < CONFIG_SYS_MAX_FLASH_BANKS; i++) {
		if (flash_info[i].flash_id != FLASH_UNKNOWN) {
			partition_size += flash_info[i].size;
		}
	}

	if (partition_size != 0) {
		partition_size -= (CONFIG_UBOOT_SIZE + \
				CONFIG_ENV_SECT_SIZE + \
				CONFIG_EPI3_SIZE);
	}
	return partition_size;
}

void show_boot_progress (int val)
{
	switch (val) {
	case BOOTSTAGE_ID_RUN_OS:
#ifdef CONFIG_PCI
		/* Put PCIe back into reset */
		x310_pcie_deinit();
#endif
		break;
	}
}

int do_reset(cmd_tbl_t *cmdtp, int flag, int argc, char * const argv[])
{
	u32 reg_val = 0;

	/* set reset bit */
	reg_val = reg_read(GPIOOUT_REG);
	reg_write(GPIOOUT_REG, reg_val & ~GPIO_MANUAL_RST);
	reg_val = reg_read(GPIOOUTEN_REG);
	reg_write(GPIOOUTEN_REG, reg_val | GPIO_MANUAL_RST);
	return 0;
}
