/*
 * fpga.c
 *
 * Copyright (c) 2012 Allied Telesis Labs
 *
 * This file is licensed under the terms of the GNU General Public License
 * version 2. This program is licensed "as is" without any warranty of any
 * kind, whether express or implied.
 */
#include <common.h>
#include <malloc.h>
#include <fpga.h>
#include <spartan3.h>
#include <asm/mpc85xx_gpio.h>

#define FPGA_CLK	(31 - 11)
#define FPGA_DATA	(31 -  9)
#define FPGA_DONE	(31 - 29)
#define FPGA_INIT	(31 - 28)
#define FPGA_PRGM	(31 - 13)
#define FPGA_PRGM_EN	(31 - 15)

extern unsigned char cfc_fpga_bin[];
extern unsigned int cfc_fpga_bin_len;

static int fpga_pre_config_fn(int cookie)
{
	/* Setting GPIOs for programming Mode */
	gpio_direction_output(FPGA_PRGM_EN, 1);
	gpio_direction_output(FPGA_PRGM, 1);
	gpio_direction_output(FPGA_CLK, 1);
	gpio_direction_output(FPGA_DATA, 0);
	gpio_direction_input(FPGA_INIT);
	gpio_direction_input(FPGA_DONE);

	/* Be sure that signal are deasserted */
	gpio_set_value(FPGA_PRGM, 1);
	gpio_set_value(FPGA_PRGM_EN, 1);

	return 0;
}

static int fpga_pgm_fn(int nassert, int nflush, int cookie)
{
	gpio_set_value(FPGA_PRGM, !nassert);

	return nassert;
}

static int fpga_init_fn(int cookie)
{
	return !gpio_get_value(FPGA_INIT);
}

static int fpga_done_fn(int cookie)
{
	return gpio_get_value(FPGA_DONE);
}

static int fpga_clk_fn(int assert_clk, int flush, int cookie)
{
	gpio_set_value(FPGA_CLK, assert_clk);

	return assert_clk;
}

static int fpga_wr_fn(int nassert_write, int flush, int cookie)
{
	gpio_set_value(FPGA_DATA, nassert_write);

	return nassert_write;
}

static int fpga_post_config_fn(int cookie)
{
	u8 val;

	udelay(100);

	// Write Enable Flash
	val = in8(FPGA_BASE + 0x40);
	val |= ((1 << 0) | (1 << 1));
	out8(FPGA_BASE + 0x40, val);

	// Un-reset eth0 PHY
	val = in8(FPGA_BASE + 0x21);
	val &= ~(1 << 0);
	out8(FPGA_BASE + 0x21, val);

	return 0;
}

Xilinx_Spartan3_Slave_Serial_fns cfc960_fpga_fns = {
	fpga_pre_config_fn,
	fpga_pgm_fn,
	fpga_clk_fn,
	fpga_init_fn,
	fpga_done_fn,
	fpga_wr_fn,
	fpga_post_config_fn,
};

Xilinx_desc fpga = XILINX_XC3S200_DESC(slave_serial, &cfc960_fpga_fns, 0);

int cfc960_fpga_init(void)
{
	int ret = -1;
#if defined(CONFIG_FIT)
	int dev;
	const void *data;
	size_t len;
	char *fit_uname_fpga = "fpga@1";
	char *desc;
	int images_noffset;
	int fpga_noffset;
#ifdef CONFIG_SPIFLASH
	void *fit_hdr = spi_firmware_get();

	if (!fit_hdr) {
		printf("Failed to load firmware blob from SPI flash\n");
		return -1;
	}
#else
	void *fit_hdr = (void *)CONFIG_FIRMWARE_BLOB_ADDR;
#endif

	if (!fit_check_format(fit_hdr)) {
		printf("Bad firmware blob format\n");
		return -1;
	}

	if (fit_get_desc(fit_hdr, 0, &desc) == 0)
		printf("Firmware: %s\n", desc);

	images_noffset = fdt_path_offset(fit_hdr, FIT_IMAGES_PATH);
	if (images_noffset < 0) {
		printf("Can't find images parent node '%s' (%s)\n",
			FIT_IMAGES_PATH, fdt_strerror(images_noffset));
		return -1;
	}

	fpga_noffset = fit_image_get_node(fit_hdr, fit_uname_fpga);
	if (fpga_noffset < 0) {
		printf("Can't find FPGA firmware node '%s' (%s)\n",
			fit_uname_fpga, fdt_strerror(fpga_noffset));
		return -1;
	}

	if (fit_image_get_data(fit_hdr, fpga_noffset, &data, &len)) {
		printf("Can't get FPGA firmware data\n");
		return -1;
	}

	fpga_init();
	dev = fpga_add(fpga_xilinx, &fpga);
	if (dev >= 0) {
		printf("Loading FPGA ... ");
		ret = fpga_load(dev, data, len);
		printf("%s\n", ret == 0 ? "OK" : "FAIL");
	}

#ifdef CONFIG_SPIFLASH
	free(fit_hdr);
#endif

#endif

	return ret;
}
