/*
 * x930.c
 *
 * Copyright 2013 Allied Telesis Labs
 *
 * Based on SBx81CFC960.c
 *
 * Copyright 2011 Freescale Semiconductor, Inc.
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#include <common.h>
#include <malloc.h>
#include <command.h>
#include <netdev.h>
#include <linux/compiler.h>
#include <asm/mmu.h>
#include <asm/processor.h>
#include <asm/cache.h>
#include <asm/immap_85xx.h>
#include <asm/fsl_law.h>
#include <asm/fsl_serdes.h>
#include <asm/fsl_portals.h>
#include <asm/fsl_liodn.h>
#include <fm_eth.h>
#include <flash.h>
#include <fdt.h>
#include <libfdt.h>
#include <epi3.h>
#include <pca953x.h>
#ifdef CONFIG_SPIFLASH
#include <spi_flash.h>
#endif
#include "../SBx8100/mpi.h"
#ifdef CONFIG_USB_STORAGE
#include <usb.h>
#include <i2c.h>
#endif

#define RESET_CONTROL1		0xffa00020

#define GPIO_DIR(pin, dir)	(dir << pin)
#define BIT(pin)		(1 << pin)

/* pin description for gpio expander (i2c addr 0x22) */
enum GPIO_EXP_PORT_0_PINS {
	TXDIS_PORT52,
	TXDIS_PORT51,
	TXDIS_PORT50,
	TXDIS_PORT49
};

static const uint pca9555_gpio_dirs =
	GPIO_DIR(TXDIS_PORT52, PCA953X_DIR_OUT) |
	GPIO_DIR(TXDIS_PORT51, PCA953X_DIR_OUT) |
	GPIO_DIR(TXDIS_PORT50, PCA953X_DIR_OUT) |
	GPIO_DIR(TXDIS_PORT49, PCA953X_DIR_OUT);

extern void pci_of_setup(void *blob, bd_t *bd);

DECLARE_GLOBAL_DATA_PTR;

int checkboard(void)
{
	struct cpu_type *cpu = gd->arch.cpu;
	ccsr_gur_t *gur = (void *)CONFIG_SYS_MPC85xx_GUTS_ADDR;
	unsigned int i;

	printf("Board: x930 (%s), ", cpu->name);

#ifdef CONFIG_PHYS_64BIT
	puts("36-bit Addressing\n");
#endif

	/*
	 * Display the RCW, so that no one gets confused as to what RCW
	 * we're actually using for this boot.
	 */
	puts("Reset Configuration Word (RCW):");
	for (i = 0; i < ARRAY_SIZE(gur->rcwsr); i++) {
		u32 rcw = in_be32(&gur->rcwsr[i]);

		if ((i % 4) == 0)
			printf("\n       %08x:", i * 4);
		printf(" %08x", rcw);
	}
	puts("\n");

	return 0;
}

int board_early_init_f(void)
{
	ccsr_gur_t *gur = (void *)(CONFIG_SYS_MPC85xx_GUTS_ADDR);

	/* board only uses the DDR_MCK0, so disable the DDR_MCK1/2/3 */
	setbits_be32(&gur->ddrclkdr, 0x001C0000);

	return 0;
}

int board_early_init_r(void)
{
	const unsigned int flashbase = CONFIG_SYS_FLASH_BASE;
	const u8 flash_esel = find_tlb_idx((void *)flashbase, 1);

	/*
	 * Remap Boot flash + PROMJET region to caching-inhibited
	 * so that flash can be erased properly.
	 */

	/* Flush d-cache and invalidate i-cache of any FLASH data */
	flush_dcache();
	invalidate_icache();

	/* invalidate existing TLB entry for flash */
	disable_tlb(flash_esel);

	set_tlb(1, flashbase, CONFIG_SYS_FLASH_BASE_PHYS,
			MAS3_SX|MAS3_SW|MAS3_SR, MAS2_I|MAS2_G,
			0, flash_esel, BOOKE_PAGESZ_256M, 1);

	set_liodns();
	setup_portals();

	return 0;
}

unsigned long get_board_sys_clk(unsigned long dummy)
{
	return 100000000;
}

static int epi3_set_macaddr(epi3Personality *epi3, int index)
{
	if(epi3) {
		switch (index) {
		case 1:
			eth_setenv_enetaddr("eth1addr", epi3->epiMacAddress);
			break;
		case 2:
			eth_setenv_enetaddr("eth2addr", epi3->epiMacAddress);
			break;
		case 3:
			eth_setenv_enetaddr("eth3addr", epi3->epiMacAddress);
			break;
		case 0:
		default:
			eth_setenv_enetaddr("ethaddr", epi3->epiMacAddress);
			break;
		}

		return 0;
	}

	return -1;
}

static void setenv_eth_addrs(void)
{
	uchar unused_eth_mac[6] = {0x00, 0x00, 0x00, 0x00, 0x00, 0x00};
#ifdef CONFIG_DIAG_EPI3
	epi3Personality *epi3 = epi3_getcurrent();

	epi3_set_macaddr(epi3, 0);
#endif

#ifdef CONFIG_DIAG_IDPROM
#endif
	eth_setenv_enetaddr("eth1addr", unused_eth_mac);
	eth_setenv_enetaddr("eth2addr", unused_eth_mac);
#ifdef CONFIG_DIAG_EPI3
	epi3_set_macaddr(epi3, 3);
#endif
	eth_setenv_enetaddr("eth4addr", unused_eth_mac);
}

static u8 get_fan_dac (void)
{
	uint dac = simple_strtoul(getenv("fancontrol"), NULL, 16);

	// ensure the DAC is set within an acceptable range
	if (dac < 0x60)
		dac = 0x60; // ~5v output on the DAC
	if (dac > 0xff)
		dac = 0xff; // 12v output on the DAC

	return dac;
}

struct i2c_addr {
	uint bus;
	uchar dev;
};

static void set_fan_cntl (u8 val)
{
	struct i2c_addr fan_controllers[] = {
			{4, 0x2c},
			{9, 0x2d},
			{8, 0x2e},
			{0, 0}
	};
	struct i2c_addr *controller;

	controller = fan_controllers;
	while (controller->bus !=0 && controller->dev !=0) {
		i2c_set_bus_num(controller->bus);
		i2c_write (controller->dev, 0x19, 1, &val, sizeof(val));
		controller++;
	}
}

static void led_7seg_init(void)
{
	struct i2c_addr led_7seg = {
			.bus=7,
			.dev = 0x20,
	};

	i2c_set_bus_num(led_7seg.bus);
	/* light up 7-segment LED */
	i2c_reg_write(led_7seg.dev, 0x1, 0x0);	/* output value */
	i2c_reg_write(led_7seg.dev, 0x3, 0x0);	/* control */
}

static void pluggable_init(void)
{
	u8 val;
	u8 mask;

	I2C_SET_BUS(PCA9555_I2C_BUS_NUM);

	/* Set the IO pins for the direction/output value */
	mask = BIT(TXDIS_PORT49) | BIT(TXDIS_PORT50) |
		   BIT(TXDIS_PORT51) | BIT(TXDIS_PORT52);

	/* configuration for TXDIS (out) */
	pca953x_set_dir(PCA9555_I2C_ADDR, cpu_to_le16(mask), cpu_to_le16(pca9555_gpio_dirs));

	/* output value for TXDIS */
	val = BIT(TXDIS_PORT49) | BIT(TXDIS_PORT50) |
		  BIT(TXDIS_PORT51) | BIT(TXDIS_PORT52);
	pca953x_set_val(PCA9555_I2C_ADDR, cpu_to_le16(mask), cpu_to_le16(val));
}

int misc_init_r(void)
{
	u8 val = 0xff;

	led_7seg_init();

	pluggable_init();

	mpi_board_init();

	setenv_eth_addrs();

	/* Enable loading of the FMAN ucode from SPI */
	if(mpi_board_attached())
		mpi_spi_flash_select(1,0);

	/* Setup fans.*/
	if (!getenv("fancontrol")) {
		/* Normally these should be forced to full speed */
		set_fan_cntl (val);
	} else {
		/* For development/test we allow slowing down the fan */
		set_fan_cntl (get_fan_dac());
	}
	return 0;
}

/* This is run after timer_interrupt is enabled */
int last_stage_init(void)
{
#ifdef CONFIG_USB_STORAGE
	usb_init();
#endif
	/* bring the switch chip out of reset, it takes some time to initialize */
	*(u8 *)RESET_CONTROL1 = 0xfe;
	return 0;
}

static int fdt_fixup_flash_partitions(void *blob, int parent)
{
	int err, node, i;
	u32 tmp[2];
	struct {
		char *name;
		u32 addr;
		u32 size;
	} paritions[] = {
		{"errlog",     0xff00000,         0x100000},
		{"user", CONFIG_UBOOT_SIZE+CONFIG_ENV_SECT_SIZE, 0xfde0000},
		{"u-boot-env", CONFIG_UBOOT_SIZE, CONFIG_ENV_SECT_SIZE},
		{"u-boot",     0,                 CONFIG_UBOOT_SIZE},
	};

	for (i = 0; i < 4; i++) {
		char part_name[32];
		snprintf(part_name, sizeof(part_name), "partition@%s",
			paritions[i].name);

		node = fdt_add_subnode(blob, parent, part_name);
		if (node < 0) {
			printf("%s: Failed to create FDT node %s %s\n",
				__func__, paritions[i].name,
				fdt_strerror(node));
			return -1;
		}

		err = fdt_setprop(blob, node, "label", paritions[i].name,
				strlen(paritions[i].name)+1);
		if (err < 0) {
			printf("%s: Failed to set \"label\" property %s %s\n",
				__func__, paritions[i].name,
				fdt_strerror(err));
			return -1;
		}

		if (paritions[i].addr == 0) {
			err = fdt_setprop(blob, node, "read-only", NULL, 0);
			if (err < 0) {
				printf("%s: Failed to set \"read-only\" "
					"property %s %s\n",
					__func__, paritions[i].name,
					fdt_strerror(err));
				return -1;
			}
		}

		tmp[0] = cpu_to_fdt32(paritions[i].addr);
		tmp[1] = cpu_to_fdt32(paritions[i].size);
		err = fdt_setprop(blob, node, "reg", tmp, sizeof(tmp));
		if (err < 0) {
			printf("%s: Failed to set \"reg\" property %s %s\n",
				__func__, paritions[i].name,
				fdt_strerror(err));
			return -1;
		}
	}

	return 0;
}

/**
 * Determines whether this fixup is required. If /flash@1
 * is present then the workaround is needed.
 */
static int fdt_fixup_flash_required(void *blob)
{
	int node;

	node = fdt_path_offset(blob, "/localbus/flash@1,0");
	if (node < 0) {
		return 0;
	}

	return 1;
}

static int fdt_fixup_flash(void *blob)
{
	int err, node;
	u32 tmp[CONFIG_SYS_MAX_FLASH_BANKS*3];
	int i;
	int bank_width = cpu_to_fdt32(2);
	int device_width = cpu_to_fdt32(2);

	if (!fdt_fixup_flash_required(blob)) {
		printf("%s: Fixup not required\n", __func__);
		return 0;
	}

	// Delete the old flash FDT entries
	node = fdt_path_offset(blob, "/localbus/flash@0,0");
	if (node < 0) {
		printf("%s: flash@0,0 : %s\n", __func__, fdt_strerror(node));
		return -1;
	}

	err =  fdt_del_node(blob, node);
	if (err < 0) {
		printf("%s: flash@0,0 %s\n", __func__, fdt_strerror(err));
		return -1;
	}

	node = fdt_path_offset(blob, "/localbus/flash@1,0");
	if (node < 0) {
		printf("%s: flash@1,0 : %s\n", __func__, fdt_strerror(node));
		return -1;
	}

	err =  fdt_del_node(blob, node);
	if (err < 0) {
		printf("%s: flash@1,0 %s\n", __func__, fdt_strerror(err));
		return -1;
	}

	// Re-create flash@0,0
	node = fdt_path_offset(blob, "/localbus");
	if (node < 0) {
		printf("%s: Could not find \"localbus\" %s",
			__func__, fdt_strerror(node));
		return -1;
	}

	node = fdt_add_subnode(blob, node, "flash@0,0");
	if (node < 0) {
		printf("%s: Could not create \"flash@0,0\" %s\n",
			__func__, fdt_strerror(node));
		return -1;
	}

	err = fdt_setprop(blob, node, "compatible", "cfi-flash",
			sizeof("cfi-flash"));
	if (err < 0) {
		printf("%s: Could not set property \"compatible\" %s",
			__func__, fdt_strerror(err));
		return -1;
	}

	for (i = 0; i < CONFIG_SYS_MAX_FLASH_BANKS; i++) {
		if (flash_info[i].flash_id == FLASH_UNKNOWN)
			continue;

		tmp[i*3] =  cpu_to_fdt32(i);
		tmp[(i*3)+1] = 0;
		tmp[(i*3)+2] = cpu_to_fdt32(flash_info[i].size);
	}

	err = fdt_setprop(blob, node, "reg", tmp, sizeof(tmp));
	if (err < 0) {
		printf("%s: Could not set property \"reg\" %s",
			__func__, fdt_strerror(err));
		return -1;
	}

	err = fdt_setprop(blob, node, "bank-width", &bank_width,
		sizeof(bank_width));
	if (err < 0) {
		printf("%s: Could not set property \"bank-width\" %s",
			__func__, fdt_strerror(err));
		return -1;
	}

	err = fdt_setprop(blob, node, "device-width", &device_width,
		sizeof(device_width));
	if (err < 0) {
		printf("%s: Could not set property \"device-width\" %s",
			__func__, fdt_strerror(err));
		return -1;
	}

	// Create partition subnodes
	err = fdt_fixup_flash_partitions(blob, node);
	if (err < 0) {
		printf("%s: Could not setup partitions",
			__func__);
		return -1;
	}

	return 0;
}

void ft_board_setup(void *blob, bd_t *bd)
{
	phys_addr_t base;
	phys_size_t size;

	ft_cpu_setup(blob, bd);

	base = getenv_bootm_low();
	size = getenv_bootm_size();

	fdt_fixup_memory(blob, (u64)base, (u64)size);

#if defined(CONFIG_HAS_FSL_DR_USB) || defined(CONFIG_HAS_FSL_MPH_USB)
	fdt_fixup_dr_usb(blob, bd);
#endif

#ifdef CONFIG_PCI
	pci_of_setup(blob, bd);
#endif

	fdt_fixup_liodn(blob);
#ifdef CONFIG_SYS_DPAA_FMAN
	fdt_fixup_fman_ethernet(blob);
#endif
	fdt_fixup_flash(blob);
}

void board_reset(void)
{
	out8(FPGA_BASE + 0x22, 0xa5);
}

#ifdef CONFIG_SPIFLASH
void *spi_firmware_get(void)
{
	int ret;
	void *blob;
	struct spi_flash *flash;

	flash = spi_flash_probe(CONFIG_ENV_SPI_BUS, CONFIG_ENV_SPI_CS,
				CONFIG_ENV_SPI_MAX_HZ, CONFIG_ENV_SPI_MODE);
	if (!flash) {
		printf("SF: probe failed\n");
		return NULL;
	} else {
		blob = calloc(1, CONFIG_FIRMWARE_BLOB_SIZE);
		ret = spi_flash_read(flash, CONFIG_FIRMWARE_BLOB_ADDR,
				CONFIG_FIRMWARE_BLOB_SIZE, blob);
		if (ret)
			printf("SF: read failed\n");
		spi_flash_free(flash);
	}

	return blob;
}
#endif

/* atl_get_env_addr()
 *
 * Get location of U-Boot environement. ENV code was hacked a little to make
 * this work. Doesn't do anything special for the CFC960.
 */
u32 atl_get_env_addr(void)
{
	return CONFIG_ENV_ADDR;
}

/* atl_get_offset_from_base()
 *
 * Not used by this product
 */
u32 atl_get_offset_from_base(void)
{
	return 0x0;
}
