/*
 * Description: PCA9665 i2c bridge driver
 * Based on: drivers/i2c/pca9564_i2c.c
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see the file COPYING, or write
 * to the Free Software Foundation, Inc.,
 * 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
 */

#include <common.h>
#include <i2c.h>
#include <pca9665.h>
#include <asm/io.h>

#define PCA_STA			(CONFIG_PCA9665_BASE + 0)
#define PCA_INDPTR		(CONFIG_PCA9665_BASE + 0)
#define PCA_DAT			(CONFIG_PCA9665_BASE + 1)
#define PCA_IND			(CONFIG_PCA9665_BASE + 2)
#define PCA_CON			(CONFIG_PCA9665_BASE + 3)

#define PCA_IND_I2CCOUNT	0x00 // byte count
#define PCA_IND_I2CADR		0x01 // own address
#define PCA_IND_I2CSCLL		0x02 // SCL LOW period
#define PCA_IND_I2CSCLH		0x03 // SCL HIGH period
#define PCA_IND_I2CPRESET	0x05 // Parallel bus reset
#define PCA_IND_I2CMODE		0x06 // I2C Bus mode

static unsigned char pca_read_reg(unsigned int reg)
{
	return readb((void *)reg);
}

static void pca_write_reg(unsigned int reg, unsigned char value)
{
	writeb(value, (void *)reg);
}

static unsigned char __maybe_unused pca_read_indirect_reg(unsigned int reg)
{
	pca_write_reg(PCA_INDPTR, reg);
	return pca_read_reg(PCA_IND);
}

static void pca_write_indirect_reg(unsigned int reg, unsigned char value)
{
	pca_write_reg(PCA_INDPTR, reg);
	pca_write_reg(PCA_IND, value);
}

static int pca_wait_busy(void)
{
	unsigned int timeout = 10000;

	while (!(pca_read_reg(PCA_CON) & PCA_CON_SI) && --timeout)
		udelay(1);

	if (timeout == 0)
		debug("I2C timeout!\n");

	debug("CON = 0x%02x, STA = 0x%02x\n", pca_read_reg(PCA_CON),
	       pca_read_reg(PCA_STA));

	return timeout ? 0 : 1;
}

/*=====================================================================*/
/*                         Public Functions                            */
/*=====================================================================*/

/*-----------------------------------------------------------------------
 * Initialization
 */
void i2c_init(int speed, int slaveaddr)
{
	pca_write_indirect_reg(PCA_IND_I2CPRESET, 0xA5);
	pca_write_indirect_reg(PCA_IND_I2CPRESET, 0x5A);
	udelay(500);

	pca_write_indirect_reg(PCA_IND_I2CMODE, 0x00); // Standard mode

	// Clock is 100kHz
	pca_write_indirect_reg(PCA_IND_I2CSCLL, 0x9D); // value of tlow
	pca_write_indirect_reg(PCA_IND_I2CSCLH, 0x86); // value of th
	udelay(500);
}

/*
 * Probe the given I2C chip address.  Returns 0 if a chip responded,
 * not 0 on failure.
 */
int i2c_probe(uchar chip)
{
	uchar res;

	pca_write_reg(PCA_CON, PCA_CON_ENSIO | PCA_CON_STA);
	pca_wait_busy();

	pca_write_reg(PCA_DAT, (chip << 1) | 1); // Chip Address + Read attribute
	pca_write_reg(PCA_CON, PCA_CON_ENSIO);

	res = pca_wait_busy();
	if ((res == 0) && (pca_read_reg(PCA_STA) == 0x48)) // Timeout & NACK has been received
		res = 1;

	pca_write_reg(PCA_CON, PCA_CON_ENSIO | PCA_CON_STO);

	return res;
}

uchar i2c_read_byte(uchar chip, uint addr, int alen)
{
	uchar data;

	pca_write_reg(PCA_CON, PCA_CON_ENSIO | PCA_CON_STA);
	pca_wait_busy();

	pca_write_reg(PCA_DAT, (chip << 1)); // Chip Address
	pca_write_reg(PCA_CON, PCA_CON_ENSIO);
	pca_wait_busy();

	if (alen > 0) {
		pca_write_reg(PCA_DAT, addr); // Register Address
		pca_write_reg(PCA_CON, PCA_CON_ENSIO);
		pca_wait_busy();
	}

	pca_write_reg(PCA_CON, PCA_CON_ENSIO | PCA_CON_STA);
	pca_wait_busy();

	pca_write_reg(PCA_DAT, (chip << 1) | 1); // Chip Address + Read attribute
	pca_write_reg(PCA_CON, PCA_CON_ENSIO);
	pca_wait_busy();

	pca_write_reg(PCA_CON, PCA_CON_ENSIO);
	pca_wait_busy();

	data = pca_read_reg(PCA_DAT);
	pca_wait_busy();

	pca_write_reg(PCA_CON, PCA_CON_ENSIO | PCA_CON_STO);

	return data;
}

int i2c_write_byte(uchar chip, uint addr, int alen, uchar data)
{
	pca_write_reg(PCA_CON, PCA_CON_ENSIO | PCA_CON_STA);
	pca_wait_busy();

	pca_write_reg(PCA_DAT, chip << 1); // Chip Address
	pca_write_reg(PCA_CON, PCA_CON_ENSIO);
	pca_wait_busy();

	if (alen > 0) {
		pca_write_reg(PCA_DAT, addr); // Register Address
		pca_write_reg(PCA_CON, PCA_CON_ENSIO);
		pca_wait_busy();
	}

	pca_write_reg(PCA_DAT, data);
	pca_write_reg(PCA_CON, PCA_CON_ENSIO);
	pca_wait_busy();

	pca_write_reg(PCA_CON, PCA_CON_STO | PCA_CON_ENSIO);

	return 0;
}

/*
 * Read/Write interface:
 *   chip:    I2C chip address, range 0..127
 *   addr:    Memory (register) address within the chip
 *   alen:    Number of bytes to use for addr (typically 1, 2 for larger
 *              memories, 0 for register type devices with only one
 *              register)
 *   buffer:  Where to read/write the data
 *   len:     How many bytes to read/write
 *
 *   Returns: 0 on success, not 0 on failure
 */
int i2c_read(uchar chip, uint addr, int alen, uchar *buffer, int len)
{
	int i;

	for (i = 0; i < len; ++i)
		buffer[i] = i2c_read_byte(chip, addr + i, alen);

	return 0;
}

int i2c_write(uchar chip, uint addr, int alen, uchar *buffer, int len)
{
	int i;

	for (i = 0; i < len; ++i)
		i2c_write_byte(chip, addr + i, alen, buffer[i]);

	return 0;
}
