/*
 * Basic I2C functions
 *
 * Copyright (c) 2003 Texas Instruments
 * Allied Telesis <www.alliedtelesis.com>
 *
 * This package is free software;  you can redistribute it and/or
 * modify it under the terms of the license found in the file
 * named COPYING that should have accompanied this file.
 *
 * THIS PACKAGE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 * Author: Jian Zhang jzhang@ti.com, Texas Instruments omap1510.c (the 
 *                                             template for this file)
 *         Graeme Campbell Allied Telesis Labs 
 *
 * Copyright (c) 2003 Wolfgang Denk, wd@denx.de
 * Rewritten to fit into the current U-Boot framework
 *
 */

#include <common.h>
#include <asm/io.h>
#include <errno.h>

typedef struct _smbus_regs {
    u32 config;
    u32 timing_config;
    u32 address;
    u32 master_fifo_control; /* 0x0c */
    u32 slave_fifo_control;  /* 0x10 */
    u32 bit_bang_control;    /* 0x14 */
    u32 pad[7];
    u32 master_cmd;          /* 0x30 */
    u32 slave_cmd;
    u32 event_enable;        /* 0x38 */
    u32 event_status;        /* 0x3c */
    u32 master_write;        /* 0x40 */
    u32 master_read;         /* 0x44 */
    u32 slave_write;         /* 0x48 */
    u32 slave_read;          /* 0x4c */
} smbus_regs;

#define CHIPCOMMONB_SMBUS		0x18038000
#define MASTER_START_BUSY_CMD	0x80000000
#define MASTER_WRITE_COMPLETE	0x80000000
#define PROT_READ_BYTE			0x800
#define PROT_WRITE_BYTE			0x600
#define RX_FIFO_FLUSH			0x80000000
#define TX_FIFO_FLUSH			0x40000000
#define SMBUS_ENABLE			0x40000000
#define TIMING_MODE_400			0x80000000
#define ENABLE_ADDRESS_0		0x80
#define MASTER_STATUS_SHIFT		25
#define MASTER_STATUS_MASK		0x7

static smbus_regs *i2c_regs = (smbus_regs *)CHIPCOMMONB_SMBUS;
static int x230_driver_initialized = 0;

static int wait_for_bb(void);

void i2c_init (int speed, int slaveadd)
{
	u32 val;

	/* clear any outstanding transactions in the master fifo */
	__raw_writel (RX_FIFO_FLUSH | TX_FIFO_FLUSH, &i2c_regs->master_fifo_control);

	/* enable smbus */
	__raw_writel (SMBUS_ENABLE, &i2c_regs->config);

	/* wait for any abort to complete and the reset to finish */
	udelay (200);

	/* turn off the 400KHz timing mode, sets to 100KHz mode */
	val = __raw_readl (&i2c_regs->timing_config);
	val &= ~TIMING_MODE_400;
	__raw_writel (val, &i2c_regs->timing_config);

	/* uboot driver works without interrupts, so turn them off */
	__raw_writel (0, &i2c_regs->event_enable);

	/* clear any outstanding interrupt events by writing back set events to clear bits */
	val = __raw_readl (&i2c_regs->event_status);
	__raw_writel (val, &i2c_regs->event_status);

	/* own address */
	if (slaveadd) {
		val = ENABLE_ADDRESS_0 | (slaveadd & 0x3f);
		__raw_writel (val, &i2c_regs->address);
	}
	udelay (100);
	x230_driver_initialized = 1;
}

static int i2c_read_byte (u8 devaddr, u8 regoffset, u8 * value)
{
	int timeout = 5;
	int ret;
	u32 val;

	/* wait until bus not busy */
	ret = wait_for_bb ();
	if (ret < 0) {
		printf("%s: timeout waiting for bus\n", __func__);
		return ret;
	}

	/* store in buffer address of remote device */
	__raw_writel (devaddr << 1, &i2c_regs->master_write);
	/* store in buffer command/offset */
	__raw_writel (regoffset, &i2c_regs->master_write);
	/* set the read bit */
	__raw_writel (devaddr << 1 | 1, &i2c_regs->master_write);
	/* send the data byte */
	__raw_writel ((u32)value | MASTER_WRITE_COMPLETE, &i2c_regs->master_write);

	/* set the master cmd register for the transaction */
	__raw_writel ((MASTER_START_BUSY_CMD | PROT_READ_BYTE | 1), &i2c_regs->master_cmd);


	/* wait for the transaction to complete */
	while (((val = __raw_readl (&i2c_regs->master_cmd)) & MASTER_START_BUSY_CMD) && timeout--) {
		udelay (100);
	}

	if ((val & MASTER_START_BUSY_CMD) == 0) {
		/* check if the transaction completed ok */
		if (((val >> MASTER_STATUS_SHIFT) & MASTER_STATUS_MASK) != 0) {
			printf("%s: Error in transaction %x", __func__, val);
			return -EREMOTEIO;
		}
	}
	else  {
		return -ETIME;
	}

	/* Transaction ok, so read the byte from the remote device */
	val = __raw_readl (&i2c_regs->master_read);
	*value = val & 0xff;
	return 0;
}

static int i2c_write_byte (u8 devaddr, u8 regoffset, u8 value)
{
	int timeout = 5;
	int ret;
	u32 val;

	/* wait until bus not busy */
	ret = wait_for_bb ();
	if (ret < 0) {
		printf("%s: timeout waiting for bus\n", __func__);
		return ret;
	}

	/* store in buffer address of remote device */
	__raw_writel (devaddr << 1, &i2c_regs->master_write);

	/* store in buffer command/offset */
	__raw_writel (regoffset, &i2c_regs->master_write);

	/* send the data byte */
	__raw_writel ((u32)value | MASTER_WRITE_COMPLETE, &i2c_regs->master_write);

	/* set the master cmd register for the transaction */
	__raw_writel ((MASTER_START_BUSY_CMD | PROT_WRITE_BYTE | 1), &i2c_regs->master_cmd);


	/* wait for the transaction to complete */
	while (((val = __raw_readl (&i2c_regs->master_cmd)) & MASTER_START_BUSY_CMD) && timeout--) {
		udelay (100);
	}

	if ((val & MASTER_START_BUSY_CMD) == 0) {
		/* check if the transaction completed ok */
		if (((val >> MASTER_STATUS_SHIFT) & MASTER_STATUS_MASK) != 0) {
			printf ("%s: Error in transaction %x", __func__, val);
			return -EREMOTEIO;
		}
	}
	else  {
		return -ETIME;
	}

	return 0;
}

int i2c_probe (uchar chip)
{

    /* if the i2c bus has not previously be intialized, then do it now */
    if (!x230_driver_initialized)
        i2c_init (CONFIG_SYS_I2C_SPEED, CONFIG_SYS_I2C_SLAVE);

    return 0;
}

int i2c_read (uchar chip, uint addr, int alen, uchar * buffer, int len)
{
	int i;

	if (alen > 1) {
		printf ("I2C read: addr len %d not supported\n", alen);
		return 1;
	}

	if (addr + len > 256) {
		printf ("I2C read: address out of range\n");
		return 1;
	}

	for (i = 0; i < len; i++) {
		if (i2c_read_byte (chip, addr + i, &buffer[i])) {
			printf ("I2C read: I/O error\n");
			i2c_init (CONFIG_SYS_I2C_SPEED, CONFIG_SYS_I2C_SLAVE);
			return 1;
		}
	}

	return 0;
}

int i2c_write (uchar chip, uint addr, int alen, uchar * buffer, int len)
{
	int i;

	if (alen > 1) {
		printf ("I2C read: addr len %d not supported\n", alen);
		return 1;
	}

	if (addr + len > 256) {
		printf ("I2C read: address out of range\n");
		return 1;
	}

	for (i = 0; i < len; i++) {
		if (i2c_write_byte (chip, addr + i, buffer[i])) {
			printf ("I2C read: I/O error\n");
			i2c_init (CONFIG_SYS_I2C_SPEED, CONFIG_SYS_I2C_SLAVE);
			return 1;
		}
	}

	return 0;
}

static int wait_for_bb (void)
{
	int timeout = 10;

	while ((__raw_readl(&i2c_regs->master_cmd) & MASTER_START_BUSY_CMD) && timeout--) {
		udelay (1000);
	}

	if (timeout <= 0) {
		printf ("timed out in wait_for_bb: MASTER_CMD=%x\n",
			__raw_readl(&i2c_regs->master_cmd));
		return -ETIME;
	}

	return 0;
}

